@@ -2,33 +2,12 @@
     models/keycache.cpp
 
     This file is part of Kleopatra, the KDE keymanager
-    Copyright (c) 2007,2008 Klarälvdalens Datakonsult AB
-                  2018 Intevation GmbH
+    SPDX-FileCopyrightText: 2007, 2008 Klarälvdalens Datakonsult AB
+    SPDX-FileCopyrightText: 2018 Intevation GmbH
+    SPDX-FileCopyrightText: 2020 g10 Code GmbH
+    SPDX-FileContributor: Ingo Klöcker <dev@ingo-kloecker.de>
 
-    Kleopatra is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
-
-    Kleopatra is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-    General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
-
-    In addition, as a special exception, the copyright holders give
-    permission to link the code of this program with any edition of
-    the Qt library by Trolltech AS, Norway (or with modified versions
-    of Qt that use the same license as Qt), and distribute linked
-    combinations including the two.  You must obey the GNU General
-    Public License in all respects for all of the code used other than
-    Qt.  If you modify this file, you may extend this exception to
-    your version of the file, but you are not obligated to do so.  If
-    you do not wish to do so, delete this exception statement from
-    your version.
+    SPDX-License-Identifier: GPL-2.0-or-later
 */
 
 #include "keycache.h"
@@ -107,7 +86,7 @@ public:
     std::vector<Key>::const_iterator find(const std::vector<Key> &keys, const char *key) const
     {
         ensureCachePopulated();
-        const std::vector<Key>::const_iterator it =
+        const auto it =
             std::lower_bound(keys.begin(), keys.end(), key, Comp<std::less>());
         if (it == keys.end() || Comp<std::equal_to>()(*it, key)) {
             return it;
@@ -120,7 +99,7 @@ public:
     std::vector<Subkey>::const_iterator find(const std::vector<Subkey> &keys, const char *key) const
     {
         ensureCachePopulated();
-        const std::vector<Subkey>::const_iterator it =
+        const auto it =
             std::lower_bound(keys.begin(), keys.end(), key, Comp<std::less>());
         if (it == keys.end() || Comp<std::equal_to>()(*it, key)) {
             return it;
@@ -145,6 +124,11 @@ public:
 
     std::vector<Key> find_mailbox(const QString &email, bool sign) const;
 
+    std::vector<Subkey>::const_iterator find_keygrip(const char *keygrip) const
+    {
+        return find<_detail::ByKeyGrip>(by.keygrip, keygrip);
+    }
+
     std::vector<Subkey>::const_iterator find_subkeyid(const char *subkeyid) const
     {
         return find<_detail::ByKeyID>(by.subkeyid, subkeyid);
@@ -247,7 +231,7 @@ private:
     struct By {
         std::vector<Key> fpr, keyid, shortkeyid, chainid;
         std::vector< std::pair<std::string, Key> > email;
-        std::vector<Subkey> subkeyid;
+        std::vector<Subkey> subkeyid, keygrip;
     } by;
     bool m_initalized;
     bool m_pgpOnly;
@@ -479,6 +463,22 @@ std::vector<Key> KeyCache::findByKeyIDOrFingerprint(const std::vector<std::strin
     return result;
 }
 
+const Subkey &KeyCache::findSubkeyByKeyGrip(const char *grip) const
+{
+    const std::vector<Subkey>::const_iterator it = d->find_keygrip(grip);
+    if (it == d->by.keygrip.end()) {
+        static const Subkey null;
+        return null;
+    } else {
+        return *it;
+    }
+}
+
+const Subkey &KeyCache::findSubkeyByKeyGrip(const std::string &grip) const
+{
+    return findSubkeyByKeyGrip(grip.c_str());
+}
+
 std::vector<Subkey> KeyCache::findSubkeysByKeyID(const std::vector<std::string> &ids) const
 {
     std::vector<std::string> sorted;
@@ -828,8 +828,20 @@ void KeyCache::remove(const Key &key)
         if (const char *keyid = subkey.keyID()) {
             const auto range = std::equal_range(d->by.subkeyid.begin(), d->by.subkeyid.end(), keyid,
                                                 _detail::ByKeyID<std::less>());
-            const auto range2 = std::equal_range(range.first, range.second, fpr, _detail::ByKeyID<std::less>());
-            d->by.subkeyid.erase(range2.first, range2.second);
+            const auto it = std::remove_if(range.first, range.second,
+                                           [fpr] (const Subkey &subkey) {
+                                               return !qstricmp(fpr, subkey.parent().primaryFingerprint());
+                                           });
+            d->by.subkeyid.erase(it, range.second);
+        }
+        if (const char *keygrip = subkey.keyGrip()) {
+            const auto range = std::equal_range(d->by.keygrip.begin(), d->by.keygrip.end(), keygrip,
+                                                _detail::ByKeyGrip<std::less>());
+            const auto it = std::remove_if(range.first, range.second,
+                                           [fpr] (const Subkey &subkey) {
+                                               return !qstricmp(fpr, subkey.parent().primaryFingerprint());
+                                           });
+            d->by.keygrip.erase(it, range.second);
         }
     }
 }
@@ -876,7 +888,7 @@ template <
     template <template <typename T> class Op> class T2
     > struct lexicographically
 {
-    typedef bool result_type;
+    using result_type = bool;
 
     template <typename U, typename V>
     bool operator()(const U &lhs, const V &rhs) const
@@ -989,18 +1001,30 @@ void KeyCache::insert(const std::vector<Key> &keys)
 
     // 6b. insert into subkey ID index:
     std::vector<Subkey> by_subkeyid;
-    by_email.reserve(subkeys.size() + d->by.subkeyid.size());
+    by_subkeyid.reserve(subkeys.size() + d->by.subkeyid.size());
     std::merge(subkeys.begin(), subkeys.end(),
                d->by.subkeyid.begin(), d->by.subkeyid.end(),
                std::back_inserter(by_subkeyid),
                _detail::ByKeyID<std::less>());
 
+    // 6c. sort by key grip
+    std::sort(subkeys.begin(), subkeys.end(), _detail::ByKeyGrip<std::less>());
+
+    // 6d. insert into subkey keygrip index:
+    std::vector<Subkey> by_keygrip;
+    by_keygrip.reserve(subkeys.size() + d->by.keygrip.size());
+    std::merge(subkeys.begin(), subkeys.end(),
+               d->by.keygrip.begin(), d->by.keygrip.end(),
+               std::back_inserter(by_keygrip),
+               _detail::ByKeyGrip<std::less>());
+
     // now commit (well, we already removed keys...)
     by_fpr.swap(d->by.fpr);
     by_keyid.swap(d->by.keyid);
     by_shortkeyid.swap(d->by.shortkeyid);
     by_email.swap(d->by.email);
     by_subkeyid.swap(d->by.subkeyid);
+    by_keygrip.swap(d->by.keygrip);
     by_chainid.swap(d->by.chainid);
 
     for (const Key &key : qAsConst(sorted)) {
@@ -1095,7 +1119,10 @@ KeyCache::RefreshKeysJob::RefreshKeysJob(KeyCache *cache, QObject *parent) : QOb
 {
 }
 
-KeyCache::RefreshKeysJob::~RefreshKeysJob() {}
+KeyCache::RefreshKeysJob::~RefreshKeysJob()
+{
+    delete d;
+}
 
 void KeyCache::RefreshKeysJob::start()
 {
