@@ -2,33 +2,10 @@
     models/useridlistmodel.cpp
 
     This file is part of Kleopatra, the KDE keymanager
-    Copyright (c) 2007 Klarälvdalens Datakonsult AB
-                  2016 Andre Heinecke <aheinecke@gnupg.org>
+    SPDX-FileCopyrightText: 2007 Klarälvdalens Datakonsult AB
+    SPDX-FileCopyrightText: 2016 Andre Heinecke <aheinecke@gnupg.org>
 
-    Kleopatra is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
-
-    Kleopatra is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-    General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
-
-    In addition, as a special exception, the copyright holders give
-    permission to link the code of this program with any edition of
-    the Qt library by Trolltech AS, Norway (or with modified versions
-    of Qt that use the same license as Qt), and distribute linked
-    combinations including the two.  You must obey the GNU General
-    Public License in all respects for all of the code used other than
-    Qt.  If you modify this file, you may extend this exception to
-    your version of the file, but you are not obligated to do so.  If
-    you do not wish to do so, delete this exception statement from
-    your version.
+    SPDX-License-Identifier: GPL-2.0-or-later
 */
 
 #include "useridlistmodel.h"
@@ -41,6 +18,16 @@
 #include <QVariant>
 #include <QIcon>
 
+#include "keycache.h"
+
+#include <gpgme++/gpgmepp_version.h>
+#if GPGMEPP_VERSION >= 0x10E00 // 1.14.0
+# define GPGME_HAS_REMARKS
+#endif
+#if GPGMEPP_VERSION >= 0x10E01 // 1.14.1
+# define GPGME_USERID_SIGNATURES_ARE_SORTABLE
+#endif
+
 using namespace GpgME;
 using namespace Kleo;
 
@@ -51,16 +38,27 @@ class UIDModelItem
     // null values. (Not null but isNull)
     //
 public:
-    explicit UIDModelItem(const UserID::Signature &sig, UIDModelItem *parentItem)
+    explicit UIDModelItem(const UserID::Signature &sig, UIDModelItem *parentItem, bool showRemarks)
     {
         mItemData  << QString::fromUtf8(sig.signerKeyID())
                    << Formatting::prettyName(sig)
                    << Formatting::prettyEMail(sig)
                    << Formatting::creationDateString(sig)
                    << Formatting::expirationDateString(sig)
-                   << Formatting::validityShort(sig);
+                   << Formatting::validityShort(sig)
+                   << (sig.isExportable() ? QStringLiteral("✓") : QString());
         mSig = sig;
         mParentItem = parentItem;
+
+        if (showRemarks && parentItem) {
+            QString lastNotation;
+            for (const auto &notation: sig.notations()) {
+                if (notation.name() && !strcmp(notation.name(), "rem@gnupg.org")) {
+                    lastNotation = QString::fromUtf8(notation.value());
+                }
+            }
+            mItemData << lastNotation;
+        }
     }
 
     explicit UIDModelItem(const UserID &uid, UIDModelItem *parentItem)
@@ -71,14 +69,19 @@ public:
     }
 
     // The root item
-    explicit UIDModelItem() : mParentItem(nullptr)
+    explicit UIDModelItem(bool showRemarks) : mParentItem(nullptr)
     {
         mItemData << i18n("ID")
                   << i18n("Name")
                   << i18n("E-Mail")
                   << i18n("Valid From")
                   << i18n("Valid Until")
-                  << i18n("Status");
+                  << i18n("Status")
+                  << i18n("Exportable");
+
+        if (showRemarks) {
+            mItemData << i18n("Tags");
+        }
     }
 
     ~UIDModelItem()
@@ -190,14 +193,17 @@ void UserIDListModel::setKey(const Key &key)
     delete mRootItem;
     mKey = key;
 
-    mRootItem = new UIDModelItem();
+    mRootItem = new UIDModelItem(mRemarksEnabled);
     for (int i = 0, ids = key.numUserIDs(); i < ids; ++i) {
         UserID uid = key.userID(i);
-        UIDModelItem *uidItem = new UIDModelItem(uid, mRootItem);
+        auto *uidItem = new UIDModelItem(uid, mRootItem);
         mRootItem->appendChild(uidItem);
-        for (int j = 0, sigs = uid.numSignatures(); j < sigs; ++j) {
-            UserID::Signature sig = uid.signature(j);
-            UIDModelItem *sigItem = new UIDModelItem(sig, uidItem);
+        std::vector<UserID::Signature> sigs = uid.signatures();
+#ifdef GPGME_USERID_SIGNATURES_ARE_SORTABLE
+        std::sort(sigs.begin(), sigs.end());
+#endif
+        for (const auto &sig : sigs) {
+            auto *sigItem = new UIDModelItem(sig, uidItem, mRemarksEnabled);
             uidItem->appendChild(sigItem);
         }
     }
@@ -237,7 +243,7 @@ int UserIDListModel::rowCount(const QModelIndex &parent) const
 QModelIndex UserIDListModel::index(int row, int column, const QModelIndex &parent) const
 {
     if (!hasIndex(row, column, parent)) {
-        return QModelIndex();
+        return {};
     }
 
     UIDModelItem *parentItem = nullptr;
@@ -259,9 +265,9 @@ QModelIndex UserIDListModel::index(int row, int column, const QModelIndex &paren
 QModelIndex UserIDListModel::parent(const QModelIndex &index) const
 {
     if (!index.isValid()) {
-        return QModelIndex();
+        return {};
     }
-    UIDModelItem *childItem = static_cast<UIDModelItem*>(index.internalPointer());
+    auto *childItem = static_cast<UIDModelItem*>(index.internalPointer());
     UIDModelItem *parentItem = childItem->parentItem();
 
     if (parentItem == mRootItem) {
@@ -292,7 +298,7 @@ QVariant UserIDListModel::data(const QModelIndex &index, int role) const
         return QVariant();
     }
 
-    UIDModelItem *item = static_cast<UIDModelItem*>(index.internalPointer());
+    auto *item = static_cast<UIDModelItem*>(index.internalPointer());
 
     if (role == Qt::ToolTipRole) {
         return item->toolTip(index.column());
@@ -305,13 +311,23 @@ QVariant UserIDListModel::data(const QModelIndex &index, int role) const
     return item->data(index.column());
 }
 
-QVector<UserID> UserIDListModel::userIDs (const QModelIndexList &indexs) const {
+UserID UserIDListModel::userID(const QModelIndex& index) const
+{
+    if (!index.isValid()) {
+        return UserID();
+    }
+    UIDModelItem *item = static_cast<UIDModelItem*>(index.internalPointer());
+    return item->uid();
+}
+
+QVector<UserID> UserIDListModel::userIDs(const QModelIndexList &indexs) const
+{
     QVector<GpgME::UserID> ret;
     for (const QModelIndex &idx : indexs) {
         if (!idx.isValid()) {
             continue;
         }
-        UIDModelItem *item = static_cast<UIDModelItem*>(idx.internalPointer());
+        auto *item = static_cast<UIDModelItem*>(idx.internalPointer());
         if (!item->uid().isNull()) {
             ret << item->uid();
         }
@@ -319,16 +335,31 @@ QVector<UserID> UserIDListModel::userIDs (const QModelIndexList &indexs) const {
     return ret;
 }
 
-QVector<UserID::Signature> UserIDListModel::signatures (const QModelIndexList &indexs) const {
+UserID::Signature UserIDListModel::signature(const QModelIndex& index) const
+{
+    if (!index.isValid()) {
+        return UserID::Signature();
+    }
+    UIDModelItem *item = static_cast<UIDModelItem*>(index.internalPointer());
+    return item->signature();
+}
+
+QVector<UserID::Signature> UserIDListModel::signatures(const QModelIndexList &indexs) const
+{
     QVector<GpgME::UserID::Signature> ret;
     for (const QModelIndex &idx : indexs) {
         if (!idx.isValid()) {
             continue;
         }
-        UIDModelItem *item = static_cast<UIDModelItem*>(idx.internalPointer());
+        auto *item = static_cast<UIDModelItem*>(idx.internalPointer());
         if (!item->signature().isNull()) {
             ret << item->signature();
         }
     }
     return ret;
 }
+
+void UserIDListModel::enableRemarks(bool value)
+{
+    mRemarksEnabled = value;
+}
